/**
 * Copyright 2010-15 Simon Andrews
 *
 *    This file is part of SeqMonk.
 *
 *    SeqMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SeqMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with SeqMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.babraham.SeqMonk.DataWriters;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

import uk.ac.babraham.SeqMonk.SeqMonkApplication;
import uk.ac.babraham.SeqMonk.SeqMonkException;
import uk.ac.babraham.SeqMonk.DataTypes.DataStore;
import uk.ac.babraham.SeqMonk.DataTypes.ProgressListener;
import uk.ac.babraham.SeqMonk.DataTypes.Probes.Probe;
import uk.ac.babraham.SeqMonk.DataTypes.Probes.ProbeList;
import uk.ac.babraham.SeqMonk.Dialogs.Cancellable;
import uk.ac.babraham.SeqMonk.Dialogs.ProgressDialog;
import uk.ac.babraham.SeqMonk.Preferences.SeqMonkPreferences;

public class BedGraphDataWriter implements Runnable, Cancellable {
	
	private DataStore data;
	private ProbeList list;
	private File file;
	private Vector<ProgressListener>listeners = new Vector<ProgressListener>();
	private boolean cancel = false;

	public BedGraphDataWriter (DataStore data, ProbeList list) {
		
		if (!data.isQuantitated()) {
			throw new IllegalArgumentException("DataStores for BEDGraph export must be quantitated");
		}
		
		this.data = data;
		this.list = list;		
		
	}
		
	
	/**
	 * Adds the progress listener.
	 * 
	 * @param l the l
	 */
	public void addProgressListener (ProgressListener l) {
		if (l != null && ! listeners.contains(l))
			listeners.add(l);
	}

	/**
	 * Removes the progress listener.
	 * 
	 * @param l the l
	 */
	public void removeProgressListener (ProgressListener l) {
		if (l != null && listeners.contains(l))
			listeners.remove(l);
	}
	
	
	public void startProcessing () {
		
		JFileChooser chooser = new JFileChooser(SeqMonkPreferences.getInstance().getSaveLocation());
		chooser.setMultiSelectionEnabled(false);
		
		chooser.setFileFilter(new FileFilter() {
			
			public String getDescription() {
				return "Text files";
			}
			
			public boolean accept(File f) {
				if (f.isDirectory() || f.getName().toLowerCase().endsWith(".txt")) {
					return true;
				}
				return false;
			}
		});
		
		int result = chooser.showSaveDialog(SeqMonkApplication.getInstance());
		if (result == JFileChooser.CANCEL_OPTION) return;

		File file = chooser.getSelectedFile();
		SeqMonkPreferences.getInstance().setLastUsedSaveLocation(file);
		
		if (file.isDirectory()) return;

		if (! file.getPath().toLowerCase().endsWith(".txt")) {
			file = new File(file.getPath()+".txt");
		}
		
		// Check if we're stepping on anyone's toes...
		if (file.exists()) {
			int answer = JOptionPane.showOptionDialog(SeqMonkApplication.getInstance(),file.getName()+" exists.  Do you want to overwrite the existing file?","Overwrite file?",0,JOptionPane.QUESTION_MESSAGE,null,new String [] {"Overwrite and Save","Cancel"},"Overwrite and Save");

			if (answer > 0) {
				return;
			}
		}
		
		this.file = file;

		addProgressListener(new ProgressDialog("Saving BEDGraph File",this));
		
		Thread t = new Thread(this);
		t.start();
	}
	
	public void run() {
		
		PrintWriter pr = null;
		
		try {
			pr = new PrintWriter(new BufferedWriter(new FileWriter(file)));
		} 
		catch (IOException ex) {
			Enumeration<ProgressListener>e = listeners.elements();
			while (e.hasMoreElements()) {
				e.nextElement().progressExceptionReceived(ex);
			}
			return;
		}
		
		/*
		 * First we need to write out the header of the file.  This sets up the preferences
		 * which are used to display the track in a web browser.  Most of the settings for
		 * this header are optional and we could add more than we do here.  The only one which
		 * is required is the type which must be set to bedGraph.
		 */
		
		pr.println("track type=bedGraph name="+(data.name().replaceAll(" ", "_"))+" visibility=full priority=20");
		
		/*
		 * Now we can add the actual data to the track
		 */
				
		Probe [] probes = list.getAllProbes();

		try {
			for (int p=0;p<probes.length;p++) {
				
				// See if we need to bail out early
				if (cancel) {
					pr.close();
					file.delete();
					Enumeration<ProgressListener>e = listeners.elements();
					while (e.hasMoreElements()) {
						e.nextElement().progressCancelled();
					}
				}
				
				// Tell the listeners how far we've got
				if (p%1000 == 0) {
					Enumeration<ProgressListener>e = listeners.elements();
					while (e.hasMoreElements()) {
						e.nextElement().progressUpdated("Saved "+p+" probes", p, probes.length);
					}
					
				}
				
				pr.println("chr"+probes[p].chromosome().name()+"\t"+probes[p].start()+"\t"+probes[p].end()+"\t"+data.getValueForProbe(probes[p]));
			}
		}
		catch (SeqMonkException ex) {
			pr.close();
			Enumeration<ProgressListener>e = listeners.elements();
			while (e.hasMoreElements()) {
				e.nextElement().progressExceptionReceived(ex);
			}
			return;
		}
		
		pr.close();
		
		Enumeration<ProgressListener>e = listeners.elements();
		while (e.hasMoreElements()) {
			e.nextElement().progressComplete("save_bedgraph", null);
		}
		
		
	}

	public void cancel() {
		cancel = true;
	}

}
