/**
 * Copyright Copyright 2010-15 Simon Andrews
 *
 *    This file is part of SeqMonk.
 *
 *    SeqMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SeqMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with SeqMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.babraham.SeqMonk.Filters;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.HashSet;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import uk.ac.babraham.SeqMonk.SeqMonkException;
import uk.ac.babraham.SeqMonk.DataTypes.DataCollection;
import uk.ac.babraham.SeqMonk.DataTypes.Genome.Chromosome;
import uk.ac.babraham.SeqMonk.DataTypes.Probes.Probe;
import uk.ac.babraham.SeqMonk.DataTypes.Probes.ProbeList;

/**
 * Filters probes based on the distance to another probe
 */
public class InterprobeDistanceFilter extends ProbeFilter {

	private Integer lowerLimit = null;
	private Integer upperLimit = null;
	private boolean allowJumps = false;

	private DistanceFilterOptionPanel optionsPanel = new DistanceFilterOptionPanel();
	

	/**
	 * Instantiates a new probe length filter with default values
	 * 
	 * @param collection The dataCollection to filter
	 * @throws SeqMonkException if the collection isn't quantitated.
	 */
	public InterprobeDistanceFilter (DataCollection collection) throws SeqMonkException {
		super(collection);
	}
	
	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#description()
	 */
	@Override
	public String description() {
		return "Filters on the length of each probe";
	}

	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#generateProbeList()
	 */
	@Override
	protected void generateProbeList() {
		
//		System.out.println("Data store size="+stores.length+" lower="+lowerLimit+" upper="+upperLimit+" type="+limitType+" chosen="+chosenNumber);

		ProbeList newList = new ProbeList(startingList,"Filtered Probes","","Length");
		allowJumps = optionsPanel.jumpBox.isSelected();
		
		// Probe lists can't handle duplicate additions so we have to manage those here.
		HashSet<Probe>addedProbes = new HashSet<Probe>();
		
		Chromosome [] chrs = collection.genome().getAllChromosomes();
		
		CHR: for (int c=0;c<chrs.length;c++) {
			progressUpdated(c, chrs.length);
			
			Probe [] probes = startingList.getProbesForChromosome(chrs[c]);
			
			P1: for (int p=0;p<probes.length;p++) {
				
				if (cancel) {
					cancel = false;
					progressCancelled();
					return;
				}
				
				// Find the first probe whose start is after this end
				int p2 = p+1;
				Probe firstProbe = null;

				for (;p2<probes.length;p2++) {
					if (probes[p2].start() > probes[p].end()) {
						firstProbe = probes[p2];
						break;
					}
				}

				// Check this probe
				if (firstProbe == null) {
					continue CHR;
				}
				
				int distance = firstProbe.start() - probes[p].end();
				
				if ((upperLimit == null || distance <= upperLimit) && (lowerLimit == null || distance >= lowerLimit)) {
					if (! addedProbes.contains(probes[p])) {
						newList.addProbe(probes[p],new Float(probes[p].length()));
						addedProbes.add(probes[p]);
					}
					if (! addedProbes.contains(probes[p2])) {
						newList.addProbe(firstProbe,new Float(probes[p2].length()));
						addedProbes.add(probes[p2]);
					}
				
					continue P1;
				}

				if (!allowJumps) {
					continue P1;
				}
				
				// If we allow jumps we can just keep on checking.
				for (;p2<probes.length;p2++) {

					distance = probes[p2].start() - probes[p].end();
					
					if ((upperLimit == null || distance <= upperLimit) && (lowerLimit == null || distance >= lowerLimit)) {
						if (! addedProbes.contains(probes[p])) {
							newList.addProbe(probes[p],new Float(probes[p].length()));
							addedProbes.add(probes[p]);
						}
						if (! addedProbes.contains(probes[p2])) {
							newList.addProbe(firstProbe,new Float(probes[p2].length()));
							addedProbes.add(probes[p2]);
						}
						continue P1;
					}
					if (upperLimit != null && distance > upperLimit) {
						continue P1;
					}
				}
				
			}

			
		}
		

		
		newList.setName("Distance between "+lowerLimit+"-"+upperLimit);
		filterFinished(newList);
				
	}

	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#getOptionsPanel()
	 */
	@Override
	public JPanel getOptionsPanel() {
		return optionsPanel;
	}

	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#hasOptionsPanel()
	 */
	@Override
	public boolean hasOptionsPanel() {
		return true;
	}

	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#isReady()
	 */
	@Override
	public boolean isReady() {
		
		if (lowerLimit == null && upperLimit == null) return false;
		
		if (lowerLimit != null && upperLimit != null && lowerLimit > upperLimit) return false;
		
		return true;
	}

	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#name()
	 */
	@Override
	public String name() {
		return "Probe Length Filter";
	}

	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#listDescription()
	 */
	@Override
	protected String listDescription() {
		StringBuffer b = new StringBuffer();
		
		b.append("Filter on probes in ");
		b.append(collection.probeSet().getActiveList().name());
		b.append(" where inter-probe distance was ");
		
		if (lowerLimit != null && upperLimit != null) {
			b.append("between ");
			b.append(lowerLimit);
			b.append(" and ");
			b.append(upperLimit);
		}
		
		else if (lowerLimit != null) {
			b.append("above ");
			b.append(lowerLimit);
		}
		else if (upperLimit != null) {
			b.append("below ");
			b.append(upperLimit);
		}
		
		if (allowJumps) {
			b.append(" allowing jumping of intervening probes");
		}
		
		return b.toString();
	}

	/* (non-Javadoc)
	 * @see uk.ac.babraham.SeqMonk.Filters.ProbeFilter#listName()
	 */
	@Override
	protected String listName() {
		StringBuffer b = new StringBuffer();
			
		b.append("Probe distance ");
		
		if (lowerLimit != null && upperLimit != null) {
			b.append("between ");
			b.append(lowerLimit);
			b.append(" and ");
			b.append(upperLimit);
		}
		
		else if (lowerLimit != null) {
			b.append("above ");
			b.append(lowerLimit);
		}
		else if (upperLimit != null) {
			b.append("below ");
			b.append(upperLimit);
		}
		
		if (allowJumps) {
			b.append(" allowing spanning");
		}
		
		return b.toString();
	}

	/**
	 * The DistanceFilterOptionPanel.
	 */
	private class DistanceFilterOptionPanel extends JPanel implements KeyListener {
			
		private JTextField lowerLimitField;
		private JTextField upperLimitField;
		private JCheckBox jumpBox;
			
		/**
		 * Instantiates a new values filter option panel.
		 */
		public DistanceFilterOptionPanel () {
			setLayout(new BorderLayout());
			
			JPanel choicePanel = new JPanel();
			choicePanel.setLayout(new BoxLayout(choicePanel,BoxLayout.Y_AXIS));


			JPanel choicePanel2 = new JPanel();
			choicePanel2.add(new JLabel("Inter-probe distance must be between "));
			lowerLimitField = new JTextField(6);
			lowerLimitField.addKeyListener(this);
			choicePanel2.add(lowerLimitField);

			choicePanel2.add(new JLabel(" and "));

			upperLimitField = new JTextField(6);
			upperLimitField.addKeyListener(this);
			choicePanel2.add(upperLimitField);
			choicePanel.add(choicePanel2);
			
			JPanel choicePanel3 = new JPanel();
			choicePanel3.add(new JLabel("Distance can jump over other probes"));
			jumpBox = new JCheckBox();
			jumpBox.setSelected(true);
			choicePanel3.add(jumpBox);
			choicePanel.add(choicePanel3);

			add(new JScrollPane(choicePanel),BorderLayout.CENTER);

		}
			
		/* (non-Javadoc)
		 * @see javax.swing.JComponent#getPreferredSize()
		 */
		public Dimension getPreferredSize () {
			return new Dimension(600,400);
		}

		/* (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
		 */
		public void keyTyped(KeyEvent arg0) {}

		/* (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
		 */
		public void keyPressed(KeyEvent ke) {}

		/* (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
		 */
		public void keyReleased(KeyEvent ke) {

			JTextField f = (JTextField)ke.getSource();

			try {
				if (f == lowerLimitField) {
					if (f.getText().length() == 0) {
						lowerLimit = null;
					}
					else {
						lowerLimit = Integer.parseInt(f.getText());
					}
				}
				else if (f == upperLimitField) {
					if (f.getText().length() == 0) {
						upperLimit = null;
					}
					else {
						upperLimit = Integer.parseInt(f.getText());
					}
				}
			}
			catch (NumberFormatException e) {
				f.setText(f.getText().substring(0,f.getText().length()-1));
			}
				
			optionsChanged();
		}
	}
}
